#include <arch/amd64/types.h>
#include <arch/amd64/vm.h>

typedef struct amd64_acpi_rsdp
{
	uint8  sig[8];
	uint8 checksum;
	uint8  oemid[6];
	uint8 revision;
	uint32 rsdt_address; 

	//uint32 length;
	//uint64 xsdt_address;
	//uint8 extended_checksum;
	//uint8 reserved[3];

} amd64_acpi_rsdp;

typedef struct amd64_acpi_rsdt_header
{
	uint8 sig[4];
	uint32 length;

	uint8 revision;
	uint8 checksum;
	uint8 oemid[6];
	uint8 oem_table_id[8];
	uint32 oem_revision;
	uint32 creator_id;
	uint32 creator_rev;
} amd64_acpi_rsdt_header;

#define ACPI_SIGNATURE "RSD PTR "

void *amd64_acpi_find()
{
struct amd64_acpi_rsdp *rsdp;
uint8 *acpi_sig; 
size_t start_address, end_address,address;
uint8 *ptr;

	start_address = KERNEL_VMA + 0xE0000;
	end_address = KERNEL_VMA + 0xFFFFF;
	acpi_sig = ACPI_SIGNATURE;

	//address = start_address;
	for(address = start_address;address < end_address; address += 16)
	{
		ptr = (uint8*) address;
		if(ptr[0] == acpi_sig[0] && ptr[1]==acpi_sig[1] &&
		   ptr[2] == acpi_sig[2] && ptr[3]==acpi_sig[3] &&
		   ptr[4] == acpi_sig[4] && ptr[5]==acpi_sig[5] &&
		   ptr[6] == acpi_sig[6] && ptr[7]==acpi_sig[7])
		{
			rsdp = (struct amd64_acpi_rsdp*)ptr;
			printk("amd64_acpi_find: found ACPI at 0x%p, revision %i.0\n",ptr,rsdp->revision+1);
			
			// TODO checksum
			return (void*)ptr;
		}

	}

	return (void*)0;
}

void amd64_acpi_debug(struct amd64_acpi_rsdp *p)
{
struct amd64_acpi_rsdt_header *header;

	printk("RSDT is at 0x%x\n",p->rsdt_address);
	header = (struct amd64_acpi_rsdt_header*) ((size_t)KERNEL_VMA + p->rsdt_address);
printk("header=%p\n",header);
printk("length=%x\n",header->length);
}

void amd64_acpi_init()
{
struct amd64_acpi_rsdp *rsdp;
	
	rsdp = (struct amd64_acpi_rsdp*)amd64_acpi_find();

	amd64_acpi_debug(rsdp);
}