
#define KERNEL_VMA 0xFFFFFFFF80000000


.section .text
.align 8
.code32

# Multiboot header for GRUB
_multiboot:
    .long 0x1BADB002
    .long 0x00010002
    .long  (0 - 0x1badb002 - 0x00010002)
    .long (_multiboot-KERNEL_VMA)
    .long (_code-KERNEL_VMA)
    .long (_data_end-KERNEL_VMA)
    .long (_end-KERNEL_VMA)
    .long (_entry-KERNEL_VMA)


.globl _entry
_entry:

	# disable interrupts
	cli

	# save multiboot stuff grub passes us
	movl %eax, (_grub_magic-KERNEL_VMA)
	movl %ebx,  (_grub_mboot-KERNEL_VMA)
	 	 
	# map the pdpt into the pml4
	movl $_boot_pdpt-KERNEL_VMA + 0x7, (_boot_pml4-KERNEL_VMA)
	
	# map pd0 into pdpt (identity mapping)
	movl $_boot_pd0-KERNEL_VMA + 0x7, (_boot_pdpt-KERNEL_VMA)
	
	
	# map the same pdpt to the end of the pml4 for our higher half	
	movl $_boot_pdpt-KERNEL_VMA + 0x7, (_boot_pml4-KERNEL_VMA)+0xFF8
		
	# map pd0 high in the address space (for higher half)
	movl $_boot_pd0-KERNEL_VMA + 0x7, (_boot_pdpt-KERNEL_VMA)+0xFF0


	# identity map the first 2 x 2MB pages	 
	movl $_boot_pd0-KERNEL_VMA, %ebx
	movl $0x87, %eax
	movl %eax, (%ebx)
	movl $0, 4(%ebx)
	
	add $8, %ebx
	addl $0x200000, %eax
	movl %eax, (%ebx)
	movl $0, 4(%ebx)

	# enable PAE
	movl %cr4, %eax
	bts $5, %eax
	movl %eax, %cr4

	# load our PML4
	movl $_boot_pml4-KERNEL_VMA,%eax
	movl %eax, %cr3

	# enable long mode
	movl $0xC0000080, %ecx
	rdmsr
	# set LME=1
	bts $8, %eax				
	wrmsr	
			
	# activate long mode by enabling paging
	movl %cr0, %eax
	bts  $31, %eax
	movl %eax, %cr0

	# fixup our GDT address to a low address
	movl	$_lmode_gdt_start-KERNEL_VMA, (_boot_gdtr-KERNEL_VMA) + 0x2
	lgdt _boot_gdtr-KERNEL_VMA
	
	# long mode here we come!!!
	ljmp $0x08, $long_mode-KERNEL_VMA
	
##################### X86-64 Long mode ######################
.code64
.align 8
long_mode:	
	
	movq $go_high, %rax
	jmp *%rax
go_high:

	cld
	// move into a temporary boot stack
	#movq $_boot_stack + 4092, %rsp
	movq $0x2000 + KERNEL_VMA, %rax
	movq %rax, %rsp
	add KERNEL_VMA, %rbp
	
	# fixup our GDT address to a high address
	movq $_lmode_gdt_start, (_boot_gdtr) + 0x2
	lgdt (_boot_gdtr)

	# pass multiboot info to main
	movq (_grub_magic), %rdi
	movq (_grub_mboot), %rsi
	call main	
1:	
	# die
	cli
	hlt
	jmp 1b





################## DATA Section ############################
.section .data

_lmode_gdt_start:
	.quad 0	
    .quad 0x0020980000000000
.L_lmode_gdt_end:

_boot_gdtr:
	.word .L_lmode_gdt_end - _lmode_gdt_start - 1
	.long 0



################## BSS Section ###############################
.section .bss
.globl _grub_magic
_grub_magic:
	.long _grub_magic

.globl _grub_mboot
_grub_mboot:
	.long _grub_mboot

.align 4096
.globl _boot_stack
_boot_stack:
	.fill 4096

.align 4096
.globl _boot_pml4
_boot_pml4:
	.fill 4096
	
.align 4096
.globl _boot_pdpt
_boot_pdpt:
	.fill 4096
	
.align 4096
.globl _boot_pd0
_boot_pd0:
	.fill 4096

.align 4096
.globl _kernel_pd0
_kernel_pd0:
	.fill 4096

	
