
#include <arch/amd64/types.h>
#include <arch/amd64/gdt.h>

struct amd64_gdtr gdtr;

// bog standard GDT
static struct amd64_gdt_desc gdt [] = {
	{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ 0xFFFF, 0, 0, 0x9A, 0xF, 0, 1, 0, 1, 0 }, /* KERNEL CS 0x8 */
	{ 0xFFFF, 0, 0, 0x92, 0xF, 0, 0, 0, 1, 0 }, /* KERNEL DS 0x10 */
	{ 0xFFFF, 0, 0, 0xF8, 0xF, 0, 1, 0, 1, 0 }, /* USER CS 0x18 */
	{ 0xFFFF, 0, 0, 0xF2, 0xF, 0, 0, 1, 1, 0 }, /* USER DS 0x20 */
	{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },			/* TSS */
};




void amd64_gdt_init()
{

	/* setup our GDT pointer */
	gdtr.limit = sizeof(gdt) - 1;
	gdtr.address = (uint64)&gdt;

	/* load the GDT */
	__asm__ volatile("lgdt %0" :: "m"(gdtr));

	/* reload the segment registers */
    __asm__("mov $0x0, %ax");
    __asm__("mov %ax, %fs");
    __asm__("mov %ax, %gs");
    __asm__("mov $0x10, %ax");
    __asm__("mov %ax, %ss");
    __asm__("mov %ax, %ds");
    __asm__("mov %ax, %es");   

}
