#include <arch/amd64/types.h>
#include <arch/amd64/vm.h>
//#include <arch/amd64/multiboot.h>

typedef struct multiboot_memory_map_t 
{
	uint32 size;
	uint32 base_addr_low,base_addr_high;
	uint32 length_low,length_high;
	uint32 type;
} multiboot_memory_map_t;

typedef struct multiboot_info
{
       uint32 flags;
       uint32 mem_lower;
       uint32 mem_upper;
       uint32 boot_device;
       uint32 cmdline;
       uint32 mods_count;
       uint32 mods_addr;

	   uint32 syms[4];
	   uint32 mmap_length;
	   uint32 mmap_addr;

} multiboot_info;


int amd64_mboot_init(uint32 mboot_addr, uint32 mboot_magic) 
{
struct multiboot_info *mbt;
struct multiboot_memory_map_t *mmap;
uint64 size;
uint8 *type;
uint64 base, length;

	if(mboot_magic != 0x2BADB002)
	{
		printk("amd64_mboot_init: bad multiboot magic: 0x%x\n",mboot_magic);
		return 0;
	}

	printk("amd64_mboot_init: good multiboot magic: 0x%x\n",mboot_magic);
	mbt = (struct multiboot_info*) ((uint64)KERNEL_VMA + mboot_addr);
	mmap = (struct multiboot_memory_map_t*)((uint64)KERNEL_VMA +mbt->mmap_addr);

	while((uint64)mmap < ((uint64)KERNEL_VMA+mbt->mmap_addr + mbt->mmap_length)) 
	{
		
		base = ((uint64)mmap->base_addr_high >> 32) + mmap->base_addr_low;
		length = ((uint64)mmap->length_high >> 32) + mmap->length_low;
		type = (mmap->type==1? "available":"reserved");
		
		printk("mmap: %p - %p (%dKb) %s\n",base,base+length,length/1024,type);
		
		size = mmap->size + sizeof(uint32);
		mmap = (struct multiboot_memory_map_t*) ((uint64)mmap + size);
	}

	return 0;
}