#include <arch/amd64/types.h>
#include <arch/amd64/idt.h>
#include <arch/amd64/gdt.h>

extern void _amd64_isr0(void);
extern void _amd64_isr1(void);
extern void _amd64_isr2(void);
extern void _amd64_isr3(void);
extern void _amd64_isr4(void);
extern void _amd64_isr5(void);
extern void _amd64_isr6(void);
extern void _amd64_isr7(void);
extern void _amd64_isr8(void);
extern void _amd64_isr9(void);
extern void _amd64_isr10(void);
extern void _amd64_isr11(void);
extern void _amd64_isr12(void);
extern void _amd64_isr13(void);
extern void _amd64_isr14(void);
extern void _amd64_isr15(void);
extern void _amd64_isr16(void);
extern void _amd64_isr17(void);
extern void _amd64_isr18(void);
extern void _amd64_isr_reserved(void);

static struct amd64_idtr idtr;
static struct amd64_idt_desc idt[255] = {0};


void amd64_idt_reserved_exception(uint64 stack_frame)
{

	printk("amd64_idt_reserved_exception() called\n");
	for(;;);
}

void amd64_idt_enter_exception(uint64 stack_frame, uint64 int_no)
{
struct amd64_stack_frame *frame;

	frame = (struct amd64_stack_frame *)stack_frame;

	printk("*** Exception #%i occured, ip=%p rsp=%p\n",frame->int_no,frame->rip,frame->rsp);
	for(;;);
}

void amd64_idt_ap_init()
{
	__asm__ volatile("lidt %0" :: "m"(idtr));
}

inline void outportb(uint16 port, uint8 val) 
{
	__asm__ volatile("outb %0, %1" : : "a" (val), "d" (port));
}



void amd64_idt_init()
{
int i;

	/* setup our IDT pointer */
	idtr.limit = sizeof(struct amd64_idt_desc)*255 - 1;
	idtr.address = (uint64)&idt;

	kmemset(idt,0,sizeof(struct amd64_idt_desc)*255);

	for(i = 0 ; i < 255 ; i++)
		amd64_idt_set_entry(i,(uint64)&_amd64_isr_reserved,KERNEL_CS,0x8E);


	amd64_idt_set_entry(0,(uint64)&_amd64_isr0,KERNEL_CS,0x8E);
	amd64_idt_set_entry(1,(uint64)&_amd64_isr1,KERNEL_CS,0x8E);
	amd64_idt_set_entry(2,(uint64)&_amd64_isr2,KERNEL_CS,0x8E);
	amd64_idt_set_entry(3,(uint64)&_amd64_isr3,KERNEL_CS,0x8E);
	amd64_idt_set_entry(4,(uint64)&_amd64_isr4,KERNEL_CS,0x8E);
	amd64_idt_set_entry(5,(uint64)&_amd64_isr5,KERNEL_CS,0x8E);
	amd64_idt_set_entry(6,(uint64)&_amd64_isr6,KERNEL_CS,0x8E);
	amd64_idt_set_entry(7,(uint64)&_amd64_isr7,KERNEL_CS,0x8E);
	amd64_idt_set_entry(8,(uint64)&_amd64_isr8,KERNEL_CS,0x8E);
	amd64_idt_set_entry(9,(uint64)&_amd64_isr9,KERNEL_CS,0x8E);
	amd64_idt_set_entry(10,(uint64)&_amd64_isr10,KERNEL_CS,0x8E);
	amd64_idt_set_entry(11,(uint64)&_amd64_isr11,KERNEL_CS,0x8E);
	amd64_idt_set_entry(12,(uint64)&_amd64_isr12,KERNEL_CS,0x8E);
	amd64_idt_set_entry(13,(uint64)&_amd64_isr13,KERNEL_CS,0x8E);
	amd64_idt_set_entry(14,(uint64)&_amd64_isr14,KERNEL_CS,0x8E);
	amd64_idt_set_entry(15,(uint64)&_amd64_isr15,KERNEL_CS,0x8E);
	amd64_idt_set_entry(16,(uint64)&_amd64_isr16,KERNEL_CS,0x8E);
	amd64_idt_set_entry(17,(uint64)&_amd64_isr17,KERNEL_CS,0x8E);
	amd64_idt_set_entry(18,(uint64)&_amd64_isr18,KERNEL_CS,0x8E);

	amd64_idt_remap_pic();

	/* load the IDT */
	__asm__ volatile("lidt %0" :: "m"(idtr));
}

void amd64_idt_set_entry(uint16 index, uint64 handler, uint16 sel, uint8 flags)
{

		idt[index].base_addr0 = (handler & 0xFFFF);
		idt[index].base_addr1 = handler >> 16;
		idt[index].base_addr2 = handler >> 32;

		idt[index].flags = flags;
		idt[index].selector = sel;
		idt[index].always0 = 0;
		idt[index].ist = 0;
		idt[index].reserved = 0;
}

void amd64_idt_remap_pic(void)
{
	/* It's a bit like PIC voodoo. */
	/* Basicallly we're moving IRQs out the way for the exceptions... */
	outportb(AMD64_MASTER_PIC, AMD64_PIC_ICW1);       
    outportb(AMD64_SLAVE_PIC, AMD64_PIC_ICW1);

    outportb(AMD64_MASTER_PIC+0x1, AMD64_PIC_MASTER_VEC);    
    outportb(AMD64_SLAVE_PIC+0x1, AMD64_PIC_SLAVE_VEC);

    outportb(AMD64_MASTER_PIC+0x1, 1<<2);     
    outportb(AMD64_SLAVE_PIC+0x1, 0x2);        

    outportb(AMD64_MASTER_PIC+0x1, AMD64_PIC_ICW4);     
    outportb(AMD64_SLAVE_PIC+0x1, AMD64_PIC_ICW4);

    outportb(AMD64_MASTER_IMR, 0xff);       
    outportb(AMD64_SLAVE_IMR, 0xff);
}