#include <arch/amd64/types.h>
#include <arch/amd64/vm.h>
#include <arch/amd64/mp.h>
#include <arch/amd64/apboot.h>
#include <alloc.h>
#include <spinlock.h>

extern uint8 apboot[];
extern uint64 _boot_pml4[];

static struct spinlock_t aptestlock = {0};

struct amd64_mp_processor_info mp_processors[16] = {0};
uint64 mp_bsp_lapic_address;
uint32 mp_bsp_idx = 0;
uint32 mp_processor_count = 0;

volatile int mp_ap_booted = 0;
volatile int offset = 0;

void *amd64_mp_detect()
{
size_t min_address, max_address;
size_t address, found_address;
uint8 *ptr;
	
	// Scan for the MP floating pointer structure

	min_address = KERNEL_VMA+0x0;
	max_address = KERNEL_VMA+0x8000;

mp_scan_again:
	
	found_address = 0;

	for(address = min_address;address < max_address; address += 16)
	{
			
		ptr = (uint8*) address;
		//printk("scanning %p\n",ptr);
		
		if(ptr[0] == '_' && ptr[1] == 'M' && ptr[2] == 'P' && ptr[3] == '_')
		{
			// Verify the checksum
			int checksum = 0;
			int i = 0 ;
			for(i=0;i<sizeof(struct mp_floating_pointer);i++)
				checksum += ptr[i];
			if((checksum&255)!=0)
				continue;				
			
			found_address = address;								
			return (void*)found_address;
		}
	}



	if(found_address == 0)
	{
		if(max_address == (KERNEL_VMA+0x8000))
		{
			min_address = KERNEL_VMA+0x9f000;
			max_address = KERNEL_VMA+0xa0000;
			goto mp_scan_again;
		}
		else
		if(max_address == (KERNEL_VMA+0xa0000))
		{
			min_address = KERNEL_VMA+0xf0000;
			max_address = KERNEL_VMA+0x100000;
			goto mp_scan_again;
		}
	
	}
	
	return (void*)0;
}

void amd64_mp_init()
{
struct mp_floating_pointer *mp_fp;
struct mp_config_table *mp_config;
int i;

	mp_fp = (struct mp_floating_pointer*) (void*)amd64_mp_detect();
	if(mp_fp == 0)
	{
		printk( "amd64_mp_init: no MP structure found\n");
		return;
	}

	mp_config = (struct mp_config_table*) ((size_t)KERNEL_VMA+mp_fp->mp_config);
	if(mp_config->mp_config_sig[0] != 'P' || mp_config->mp_config_sig[1] != 'C' || mp_config->mp_config_sig[2] != 'M' ||
	   mp_config->mp_config_sig[3] != 'P')
	{
		printk("amd64_mp_init: bad signature in MP config table\n");
		return;
	}

	switch (mp_config->mp_config_revision)
	{
	case 1	:	printk( "amd64_mp_init: MP Specification rev. 1.1\n"); break;
	case 4  :	printk( "amd64_mp_init: MP Specification rev. 1.4\n"); break;
	default	:	printk( "amd64_mp_init: MP Specification rev. unknown\n");break;
	}
	

	amd64_mp_parse_config(mp_config);
	
	i = 0;
	while(i < mp_processor_count )
	{
		printk("amd64_mp_init: cpu %i, apic_id=%i, %s\n",mp_processors[i].cpu_no,mp_processors[i].apic_id,
		mp_processors[i].cpu_info&MP_CPU_INFO_BOOTSTRAP? "BSP":"AP");

		i++;
	}

	printk("amd64_mp_init: bsp apic id %i\n",mp_processors[mp_bsp_idx].apic_id);

	// Initalize the BSP APIC
	amd64_apic_write(mp_bsp_lapic_address,MP_LAPIC_SPURIOUS_INT_REG, 0x00000100 | 0x7f);
	amd64_apic_write(mp_bsp_lapic_address,MP_LAPIC_TASK_PRI_REG,0xff);

	struct spinlock_t *pp;
	pp = &aptestlock;
	spinlock_init(pp);

	//amd64_mp_spinup_ap(1);
	
	offset=0;
	for(i = 0 ; i < mp_processor_count;i++)
	{
		// spinup all APs
		if(!(mp_processors[i].cpu_info&MP_CPU_INFO_BOOTSTRAP))
		{
			offset += 2;
		
			mp_ap_booted = 0;
			amd64_mp_spinup_ap(i);

			while(mp_ap_booted == 0)
				;

		}
	}

	
}

void amd64_mp_test_func(void)
{
uint8 *p = (uint8*)((uint64)KERNEL_VMA + 0xb8000);
uint8 *pp;
int i,j;
	
	pp = (uint8*)&aptestlock;
	 
again:
	 spinlock_lock(pp);
	 for(i = 0 ; i < 10000;i++)
		 p[11]++;

	 spinlock_unlock(pp);
	 for(j = 0 ; j < 1000 ; j++)
		__asm__("nop");
		
	 goto again;

}

void amd64_mp_ap_entry_point(void)
{
size_t *new_stack;

printk("ok\n");
	// load an IDT
	amd64_idt_ap_init();

	new_stack = (size_t*) mm_page_alloc();
	if(!new_stack)
	{
		printk("amd64_mp_ap_entry(): unable to alloc new stack!!");
		for(;;) 
			__asm__("cli; hlt"); // fix me
	}

	kmemset(new_stack,0,PAGESIZE);
	// switch to this new stack
	__asm__("movq %0, %%rsp\n\t"::"r"((uint64)new_stack + ((PAGESIZE)-100)):"rcx");
	
	amd64_mp_ap_ready();
}

void amd64_mp_ap_ready()
{
uint8 *p = (uint8*)((uint64)0xB8000 + KERNEL_PHYSICAL_AREA_START);
int i, off,j;

//for(;;);
//printk("ok\n");	
	uint8 *pp = (uint8*)&aptestlock;
	

	off = offset;
	mp_ap_booted = 1;
again:

//for(;;);
	spinlock_lock(pp);
	for(i = 0 ; i < 10000 ; i ++)
	{
		for(j = 0 ; j < 5 ; j++)
			__asm__("nop");
			
		p[1+off]+=off;
		p[off] ++;
	}
	spinlock_unlock(&aptestlock.holder);

	for(j = 0 ; j < 1000 ; j++)
		__asm__("nop");
	goto again;

	for(;;);
}

void amd64_mp_spinup_ap(int processor_idx)
{
uint8 *code_ptr;
uint32 *pml4_address;
uint64 *entry_point;

	printk("amd64_mp_spinup_ap: spinning up cpu #%i......",processor_idx);

	code_ptr = (uint8*)((uint64)KERNEL_VMA + 0x1000);

	// copy bootstrap code to 0x1000 physical
	kmemcpy(code_ptr,apboot,apboot_size);

	// give AP its pml4 address
	pml4_address = (uint32*)((uint64)KERNEL_VMA + 0x1000 - 8);
	*pml4_address = (uint32) ((uint32)_boot_pml4 - KERNEL_VMA);

	// give AP the kernel entry point
	entry_point = (uint64*)((uint64)KERNEL_VMA + 0x1000 - 16);
	*entry_point = (uint64) &amd64_mp_ap_entry_point;

	// Initalize the BSP APIC
	amd64_apic_write(mp_bsp_lapic_address,MP_LAPIC_SPURIOUS_INT_REG, 0x00000100 | 0x7f);
	amd64_apic_write(mp_bsp_lapic_address,MP_LAPIC_TASK_PRI_REG,0xff);

	// send INIT
	amd64_apic_ipi(mp_bsp_lapic_address,mp_processors[processor_idx].apic_id,0x500,0x01);

	int j;
	for(j = 0 ; j < 10000 ; j++) 
		__asm__("nop");


	// Startt the AP executing the code at 0x1000 ......
	amd64_apic_ipi(mp_bsp_lapic_address,mp_processors[processor_idx].apic_id,0x600,0x01);
	for(j = 0 ; j < 10000 ; j++) 
		__asm__("nop");

	amd64_apic_ipi(mp_bsp_lapic_address,mp_processors[processor_idx].apic_id,0x600,0x01);

}

void amd64_mp_parse_config(struct mp_config_table *config)
{
uint32 mp_lapicAddress;
uint8 *p;
uint8 entry_type;
int i;

	p = (void*)((size_t)config + sizeof(struct mp_config_table));

	mp_lapicAddress = (uint32)config->local_apic;

	mp_bsp_lapic_address = KERNEL_PHYSICAL_AREA_START + mp_lapicAddress;
	printk("amd64_mp_parse_config: LAPIC is at 0x%p\n",mp_lapicAddress);

	for(i = 0;i<config->mp_config_entry_count;i++)
	{

		entry_type = (uint8)p[0];
		switch(entry_type)
		{
			case MP_CONFIG_PROCESSOR:
			{
				struct mp_processor_entry *mp_proc = (void*)p;

				//printk("found a processor\n");
				//if(mp_proc->cpu_info&MP_CPU_INFO_ENABLED)
				//{
				
				mp_processors[mp_processor_count].cpu_no = mp_processor_count;
				mp_processors[mp_processor_count].apic_id = mp_proc->local_apic_id;
				mp_processors[mp_processor_count].cpu_info = mp_proc->cpu_info;
				mp_processor_count ++;			

				// Remember which processor is the BSP
				if(mp_proc->cpu_info & MP_CPU_INFO_BOOTSTRAP)
					mp_bsp_idx = i;
				//}
				//mp_processorCount ++;
		
				p += MP_PROC_LEN;
				break;
			}

			case MP_CONFIG_BUS:
			{
				
				
				p += MP_BUS_LEN;
				
				break;
			}
			
			case MP_CONFIG_IO_APIC:
			{
				//mp_ioapicCount++;				

				p += MP_IOAPIC_LEN;
				
				break;
			}

			default:
				p += 8;
				//printk("amd64_: unknown MP entry\n");
		}

		
	}

	return;
}
