#include <arch/amd64/types.h>
#include <arch/amd64/vm.h>

extern volatile uint64 _boot_pml4[];
extern volatile uint64 _boot_pdpt[];
extern volatile uint64 _boot_pd0[];
extern volatile uint64 _kernel_pd0[];
extern volatile uint64 _boot_pd4;

extern void _end(void);

static uint64 first_usable_address,last_usable_address;

#define VM_PG_MASK	0x000FFFFFFFFFF000
#define VM_PTBL_SIZE	0x200000
#define VM_PDIR_SIZE	0x40000000
#define VM_PDP_SIZE	0x8000000000

int amd64_get_pml4_no(uint64 address)
{
	return (int) (address >> 39) & 511;
}

int amd64_get_pdpt_no(uint64 address)
{
	return (int) (address >> 30) & 511;
}

int amd64_get_pd_no(uint64 address)
{
	return (int) (address >> 21) & 511;
}

int amd64_get_pte_no(uint64 address)
{
	return (int) (address >> 12) & 511;
}

int amd64_invlpg(uint64 addr)
{
	__asm__ volatile("invlpg (%0)" : : "r" (addr) : "memory");

}

int amd64_paging_debug()
{
uint64 cr3;
uint64 *val;
	__asm__ volatile("movq %%cr3 ,%0 ": "=r" (cr3));

	printk("cr3=0x%p\n_boot_pml4 @ %p\n",cr3,((uint64)_boot_pml4)-KERNEL_VMA);
	
	printk("_boot_pml4[0]=0x%p\n",_boot_pml4[0]);
	printk("_boot_pdpt @ 0x%p\n",((uint64)_boot_pdpt)-KERNEL_VMA);	
	printk("_boot_pdpt[0]=0x%p\n",_boot_pdpt[0]);
}

void amd64_paging_map_lowheap()
{
uint64 lowest_addr;

//	lowest_addr = (uint64)&end;
	
//	lowest_addr -= KERNEL_VMA;
//	lowest_addr &= VM_PG_MASK;
}

uint64 amd64_get_low_addr()
{
	return first_usable_address;
}

uint64 amd64_get_high_addr()
{
	return last_usable_address;
}

void amd64_paging_map(uint64 addr)
{
uint64 lowest_addr;
uint64 num_page_tables;
uint64 num_pdpt;

uint64 *first_pdpt, *current_pdpt;
uint64 *first_pd, *current_pd;

uint64 i,j,k;
uint64 address,highest_addr;

int base_pml4, base_pdpt;

	highest_addr = addr;
	if(highest_addr < 0xffffffff)
		highest_addr = 0xffffffff;
	base_pml4 = amd64_get_pml4_no(KERNEL_PHYSICAL_AREA_START);
	base_pdpt = amd64_get_pdpt_no(KERNEL_PHYSICAL_AREA_START);

	printk("base_pml4=%i base_pdpt=%i\n",base_pml4,base_pdpt);
	
	// calculate where we can put the first PDPT
	lowest_addr = ((uint64)&_end)-KERNEL_VMA + PAGESIZE;
	lowest_addr &= PAGE_MASK;
	printk("lowest_addr=0x%p, highest_addr=0x%p\n",lowest_addr,highest_addr);


	num_page_tables = highest_addr / VM_PDIR_SIZE;
	if(highest_addr & (VM_PDIR_SIZE-1))
		num_page_tables ++;

	num_pdpt = highest_addr / (VM_PDIR_SIZE*512L);
	if(highest_addr & ((VM_PDIR_SIZE*512L)-1))
		num_pdpt ++;

	first_usable_address = lowest_addr + (num_page_tables*PAGESIZE) + (num_pdpt*PAGESIZE);
	last_usable_address = addr;
	printk("user available RAM 0x%p-0x%p\n",first_usable_address,last_usable_address);

	printk("amd64_paging_map: page tables using %d KB\n",((num_page_tables*PAGESIZE)+(num_pdpt*PAGESIZE)) / 1024);
	first_pdpt = (uint64*)((uint64)lowest_addr + KERNEL_VMA);
	first_pd = (uint64*)((uint64)lowest_addr +KERNEL_VMA + (num_pdpt*PAGESIZE));
	
	printk("first_pdpt=0x%p, first_pd=0x%p\n",first_pdpt,first_pd);

	address = 0;
	for(i = 0 ; i < num_pdpt ; i++)
	{
		current_pdpt = (uint64*)((uint64)first_pdpt + (i*PAGESIZE));
		printk("current=%p\n",current_pdpt);
		//for(;;);
		for(j = 0 ; j < num_page_tables; j++)
		{
			current_pd = (uint64*)((uint64)first_pd + (j*PAGESIZE));
			for(k = 0 ; k < PAGESIZE/sizeof(uint64); k++)
				current_pd[k] = 0;
			
			for(k=0; k < PAGESIZE/sizeof(uint64); k++)
			{
				current_pd[k] = (uint64)address + 0x87; 
				address += 0x200000;
				if(address > highest_addr)
					break;
			}

			current_pdpt[base_pdpt+j] = (uint64) ((uint64)current_pd-KERNEL_VMA) + 0x7;
//			printk("pdpt #%i 0x%p -> 0x%p\n",j,current_pdpt[j],current_pd);
		}


		_boot_pml4[base_pml4+i] = (uint64)((uint64)current_pdpt-KERNEL_VMA) + 0x7;
//		printk("just set %p %p\n",_boot_pml4[base_pml4+i],current_pdpt);
		//for(;;);
		
	}

//	printk("last address was %x\n",address);
	printk("num_page_tables=%i\n",num_page_tables);
	printk("num_pdpt=%i\n",num_pdpt);
}

void amd64_debug()
{
int i ;

//for(i = 0 ; i < 512; i++)
//if(_boot_pml4[i] != 0)
//	printk("pml4[%i]=%p\n",i,_boot_pml4[i]);
}

void amd64_paging_init(uint64 lowest_addr, uint64 highest_addr)
{
uint64 address;
int i;

	address = 0;
	

	//_boot_pml4[511] = (uint64)((uint64)(_boot_pdpt-KERNEL_VMA) + 0x7);

	//Map the first 1GB with 2MB pages to 0xFFFFFFFF80000000 ->
	//This is required to be able to build more pagetables below
	for(i = 0 ; i < 512 ; i++)
	{
		_kernel_pd0[i] = address + 0x87;
		address += 0x20000;
	}

		/* reload CR3, flush TLB */
	__asm__ volatile ("movq %0, %%cr3" :: "r" (((uint64)_boot_pml4)-KERNEL_VMA): "memory");


	amd64_paging_map(0xfff0000);

	/* Remove the identity mapping used when we first enabled long mode */
	//_boot_pdpt[0] = 0;
	//_boot_pml4[0] = 0;
		
	/* reload CR3, flush TLB */
	__asm__ volatile ("movq %0, %%cr3" :: "r" (((uint64)_boot_pml4)-KERNEL_VMA): "memory");

	//int val = amd64_get_pde_no(KERNEL_VMA);
//	uint32 val = ((0xC0000000000) >> 39) & 511;
//	printk("val=%p %u\n",val,val);
}