#ifndef __INCLUDE_ARCH_AMD64_MP_H
#define __INCLUDE_ARCH_AMD64_MP_H

#include <arch/amd64/types.h>

#define MP_CONFIG_PROCESSOR 0
#define MP_CONFIG_BUS 1
#define MP_CONFIG_IO_APIC 2
#define MP_CONFIG_IO_INT_ASSIGNMENT 3
#define MP_CONFIG_LOCAL_INT_ASSIGNMENT 4

#define MP_PROC_LEN 20
#define MP_BUS_LEN 8
#define MP_IOAPIC_LEN 8
#define MP_IO_INT_ASSIGNMENT_LEN 8
#define MP_LOCAL_INT_ASSIGNMENT_LEN 8

#define	MP_CPU_INFO_ENABLED		1
#define	MP_CPU_INFO_BOOTSTRAP	2



#define MP_LAPIC_DEFAULT 0xfee00000
#define MP_IOAPIC_DEFAULT 0xfec00000

#define MP_LAPIC_SPURIOUS_INT_REG 0xF0
#define MP_LAPIC_TASK_PRI_REG 0x80
#define MP_LAPIC_ICR_LOW_REG 0x300
#define MP_LAPIC_ICR_HIGH_REG 0x310

#define MP_ICR_TRIGGER 0x8000
#define MP_ICR_LEVEL 0x4000
#define MP_ICR_STARTUP (0x6 << 8)

struct amd64_mp_processor_info
{
	uint8 cpu_no;
	uint8 apic_id;
	uint8 cpu_info;
};


struct mp_floating_pointer
{
	uint8 mp_signature[4]; // should be _MP_
	
	uint32 mp_config;
	uint8 mp_length;

	uint8 mp_version;
	uint8 mp_features1;
	uint8 mp_features2;
	uint8 mp_features3;
	uint8 mp_features4;
	uint8 mp_features5;
} __attribute__((packed));

struct mp_config_table
{
	uint8 mp_config_sig[4];
	uint16 mp_config_len;

	uint8 mp_config_revision;
	uint8 mp_config_checksum;
	
	uint8 oem_id[8];
	uint8 product_id[12];	
	uint32 oem_table_pointer;
	uint16 oem_table_size;

	uint16 mp_config_entry_count;
	
	uint32 local_apic;

	uint16 mp_config_ext_len;
	uint8 mp_config_ext_checksum;

	uint8 dummy; // this isn't in the spec!?!
} __attribute__((packed));

struct mp_processor_entry
{
	uint8 entry_type;
	uint8 local_apic_id;
	uint8 local_apic_version;
	uint8 cpu_info;
	uint8 cpu_signature[4];
	uint8 cpu_features[4];
} __attribute__((packed));

	
#endif
