
#include <types.h>
#include <spinlock.h>
#include <alloc.h>
#include <arch/amd64/vm.h>

#define panic printk

struct spinlock_t kmalloc_lock;
struct mm_page_t *mm_free_page_stack;
//struct mm_page_t *mm_available_stacks[5];
struct mm_page_t *mm_first_page;

size_t mm_lowest_addr;
size_t mm_pages_available;
size_t mm_bytes_allocated = 0;
size_t *mm_kheap_start;

void mm_page_push_free_stack(struct mm_page_t *pp)
{

	if(pp->next)
		panic("mm_push_stack() page is already on a stack page=%p\n",pp);

	pp->next = (struct mm_page_t*)mm_free_page_stack;
	mm_free_page_stack = (struct mm_page_t*) pp;
}

struct mm_page_t *mm_page_pop_free_stack()
{
struct mm_page_t *pp;

again:
	if(mm_free_page_stack)
	{
	
		pp = (struct mm_page_t*)mm_free_page_stack;
		mm_free_page_stack = (struct mm_page_t*)pp->next;


		if(pp->status != MM_PAGE_STATUS_AVAIL)
		{
		//	printk("skipping");
			goto again;
		}
	}
	else
		return NULL;
	
	pp->next = NULL;
	return (mm_page_t*)pp;
}

struct mm_page_t *mm_pop_stack(struct mm_page_t **stack)
{
struct mm_page_t *p;

	if(*stack)
	{
		p = (struct mm_page_t*)*stack;
		*stack = (struct mm_page_t *)p->next;
	}
	else
		return NULL;

	p->next = NULL;
	return p;
}

void mm_push_stack(struct mm_page_t **stack, struct mm_page_t *p)
{


	if(p->next)
		panic("mm_push_stack() page is already on a stack\n");
	
	p->next = (struct mm_page_t*)*stack;
	*stack = (struct mm_page_t*)p;
}






void mm_page_init(size_t first_address, size_t last_address)
{
struct mm_page_t *p;
size_t pages_available;
int i;


		mm_bytes_allocated = 0;
		mm_free_page_stack = NULL;

		//for(i = 0 ; i < 5 ; i ++)
		//	mm_available_stacks[i] = NULL;
	
		pages_available = (last_address - first_address) / PAGESIZE;

		mm_lowest_addr = first_address + (pages_available * sizeof(struct mm_page_t));
		mm_lowest_addr = (mm_lowest_addr + PAGESIZE) & PAGE_MASK;
		mm_pages_available = pages_available;

		printk("kmm: pages_avail=%d lowest_addr=%p\n",pages_available,mm_lowest_addr);


		mm_first_page = (struct mm_page_t*)((size_t)first_address+KERNEL_PHYSICAL_AREA_START);
        for(i = pages_available ; i >= 0 ; i --)
		{
			
			mm_first_page[i].physical_address = (mm_lowest_addr + (i*PAGESIZE));
			mm_first_page[i].next = NULL;
			mm_first_page[i].status = MM_PAGE_STATUS_AVAIL;

			p = (struct mm_page_t*)&mm_first_page[i];

			mm_page_push_free_stack((struct mm_page_t*)p);
		}

		spinlock_init(&kmalloc_lock);

}

void *mm_page_alloc_cont(int no)
{
struct mm_page_t *page, *prev, *next;
int i, found,current;
void *addr;

	spinlock_lock(&kmalloc_lock);
printk("mm_page_alloc_cont(%i)\n",no);

	found = current = i = 0;
	while(current < mm_pages_available)
	{
		page = (struct mm_page_t*)&mm_first_page[current];	
		if(page->status == MM_PAGE_STATUS_AVAIL)
		{
			found = current;
			while(page->status == MM_PAGE_STATUS_AVAIL && i < no)
			{
				current ++;
				i++;
				page = (struct mm_page_t*)&mm_first_page[current];
			}
		}
	
		if(i == no)
		{
			printk("mm_page_alloc_cont() found %i pages starting at #%i index\n",i,found);
			for(i=0;i<no;i++)
			{
				page = (struct mm_page_t*)&mm_first_page[i+found];
				if(page->status != MM_PAGE_STATUS_AVAIL)
					printk("mm_page_alloc_cont() wtf status=%i\n",page->status);

				page->status = MM_PAGE_STATUS_CHAINED;
				
				if(i==no)
					page->status == MM_PAGE_STATUS_CHAINED_LAST;

				
			}

			
			page = (struct mm_page_t*)&mm_first_page[found];
			addr = (void*)((size_t)page->physical_address + KERNEL_PHYSICAL_AREA_START);
			//printk("cont returning %p found=%i  addr=0x%p\n",addr,found,page);
			spinlock_unlock(&kmalloc_lock);
			return (void*)addr;
		}
		else
		{
			found = i = 0;
			current ++;
		}
		
	}

	spinlock_unlock(&kmalloc_lock);
	return (void*)0;
}


void *mm_page_alloc()
{
struct mm_page_t *page;
size_t addr;
	
//printk("mm_page_alloc()\n");
	spinlock_lock(&kmalloc_lock);

	page = (struct mm_page_t*) mm_page_pop_free_stack();
	if(!page)
	{
		spinlock_unlock(&kmalloc_lock);
		return NULL;
	}

	if(page->status != MM_PAGE_STATUS_AVAIL || page->next)
		printk("mm_page_alloc() WARNING page status=%i, next=%p\n",page->status,page->next);

	page->status = MM_PAGE_STATUS_ALLOCATED;

	addr = (size_t) page->physical_address + KERNEL_PHYSICAL_AREA_START;
	//printk("mm_page_alloc() returning %p\n",addr);
	spinlock_unlock(&kmalloc_lock);
	return (void*)(addr);
}

void mm_page_free_cont(size_t index)
{
struct mm_page_t *page;
size_t current;

printk("mm_page_free_cont(%i)\n",index);
	// kmalloc lock should already be held...

	current = index;
	page = (struct mm_page_t*)&mm_first_page[current];

	if(page->status != MM_PAGE_STATUS_CHAINED && page->status != MM_PAGE_STATUS_CHAINED_LAST)
	{
		printk("mm_page_free_cont() WARNING: first page not chained!\n");
		return;
	}

	while(page->status == MM_PAGE_STATUS_CHAINED)
	{
	
		page->status = MM_PAGE_STATUS_AVAIL;
		
		// If this page has been popped off, stick it back on now
		if(page->next == NULL)
			mm_page_push_free_stack(page);

		current ++;
		page = (struct mm_page_t*)&mm_first_page[current];
	}

	
	printk("mm_page_free_cont() freed %i pages\n",current);
	return;
}


void mm_page_free(void *addr)
{
struct mm_page_t *p;
size_t offset,index;

printk("mm_page_free() %p\n",addr);
	if(!addr)
		panic("kfree() addr==NULL");

  	offset = ((size_t)addr - (mm_lowest_addr+KERNEL_PHYSICAL_AREA_START));
	index = (offset & PAGE_MASK) / PAGESIZE;

	if(index > mm_pages_available)
	{
		printk("mm_page_free() WARNING: bad address 0x%p, index=%i\n",addr,index);
		return ;
	}
	
//printk("mm_page_free() mm_lowest_addr=%p offset=%p index=%i\n",mm_lowest_addr,offset,index);

	spinlock_lock(&kmalloc_lock);

	p = (struct mm_page_t *)&mm_first_page[index];
	if(p->status != MM_PAGE_STATUS_ALLOCATED && p->status != MM_PAGE_STATUS_CHAINED)
	{
		printk("mm_page_free() WARNING: bad address 0x%p, status=%i\n",addr,p->status);
		spinlock_unlock(&kmalloc_lock);
		return;
	}

	if(p->status == MM_PAGE_STATUS_ALLOCATED)
	{
		mm_page_push_free_stack(p);
		spinlock_unlock(&kmalloc_lock);	
		return;
	}
	else
	{
		mm_page_free_cont(index);
		spinlock_unlock(&kmalloc_lock);
	}

}

